/*--------------------------------------------------------------------------*/
/* Thermocouple Temperature Meter                                           */
/*--------------------------------------------------------------------------*/
/* This program is opened under license policy of following trems.
/
/  Copyright (C) 2012, ChaN, all right reserved.
/
/ * This program is a free software and there is NO WARRANTY.
/ * No restriction on use. You can use, modify and redistribute it for
/   personal, non-profit or commercial use UNDER YOUR RESPONSIBILITY.
/ * Redistributions of source code must retain the above copyright notice.
/
/---------------------------------------------------------------------------*/


#include <stdint.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include "xitoa.h"


FUSES = {0x52, 0xD9, 0xFD};	/* Fuse values: {Low, High, Ext} */
/* This is the fuse settings for this project. The fuse data will be output
   into the hex file with program code. However some old flash programmers
   do not support this sort of hex files. If it is the case, use these values
   to program the fuse bits.
*/


#define	T_IVT	150		/* Trigger interval: 300 = 1sec */
#define V_LVH	1.2		/* Normal voltage */
#define V_LVD	1.1		/* LVD voltage */
#define V_SDN	1.0		/* Shutdown voltage */
#define VREF	3.0		/* VREF voltage */

#define	BTN_PWR	(!(PINA & _BV(6)))


/*--------------------------------------------------------------------------*/
/* Work Area                                                                */
/*--------------------------------------------------------------------------*/

volatile uint8_t Lcd[3];	/* Lcd segment buffer */
/*         bit7   bit6  bit5  bit4  bit3  bit2  bit1  bit0 */
/* Lcd[0]: g1000  a100  b100  c100  d100  e100  f100  g100 */
/* Lcd[1]: LOBAT  a10   b10   c10   d10   e10   f10   g10  */
/* Lcd[2]: bc1000 a1    b1    c1    d1    e1    f1    g1   */


/* Software UART transmission fifo */
typedef struct {
	uint8_t ri, wi, ct;
	uint8_t buff[16];
} FIFO;
volatile FIFO TxFifo;

/* Status flags, timer */
volatile uint8_t Flag;	/* b1:LVD, b0:Trigger */


/*---------------------------------------------*/
/* 300Hz Interval Timer Process                */
/*---------------------------------------------*/

ISR(TIMER0_COMPA_vect)
{
	/* LCD glass driving */
	{
		uint8_t f;

		f = PINA & _BV(0) ? 0 : 0xFF;
		PORTC = f ^ Lcd[0];
		PORTD = f ^ Lcd[1];
		PORTB = f ^ Lcd[2];
		PINA = _BV(0);
	}

	/* Software UART transmission */
	{
		static uint8_t bc, td;
		uint8_t n;

		if (bc) {	/* Transmission in progress */
			if (td & 1) PORTA |= _BV(1);	/* Space */
			else		PORTA &= ~_BV(1);	/* Mark */
			td >>= 1;
			if (++bc == 11) bc = 0;
		} else {	/* Transmission idle */
			n = TxFifo.ct;
			if (n) {	/* Start transmission if tx data is available */
				TxFifo.ct = n - 1;
				n = TxFifo.ri;
				td = ~TxFifo.buff[n];
				TxFifo.ri = (n + 1) % sizeof TxFifo.buff;
				bc = 1;
				PORTA |= _BV(1);	/* Space (start bit) */
			}
		}
	}

	/* System controls */
	{
		static uint16_t iv;

		if (++iv >= T_IVT) {
			iv = 0;
			Flag |= 1;
		}
	}
}



/*---------------------------------------------*/
/* Put a character for UART                    */
/*---------------------------------------------*/

void suart_putc (
	uint8_t c
)
{
	uint8_t i;


	while (TxFifo.ct >= sizeof TxFifo.buff) ;

	i = TxFifo.wi;
	TxFifo.buff[i] = c;
	TxFifo.wi = (i + 1) % sizeof TxFifo.buff;
	cli();
	TxFifo.ct++;
	sei();
}



/*---------------------------------------------*/
/* Check fuel level                            */
/*---------------------------------------------*/

int check_fuel (void)	/* 1:Ok, 0:Empty */
{
	uint16_t ad;


	/* Get ADC value on the AIN7 pin with VREF=Vcc */
	ADMUX = 7;
	ADCSRA = _BV(ADEN)|_BV(ADSC)|_BV(ADIF)|0b000;
	while (!(ADCSRA & _BV(ADIF))) ;
	ad = ADC;
	ADCSRA = 0;

	if (ad >= (uint16_t)(1024 * V_LVH / VREF)) Flag &= ~2;
	if (ad < (uint16_t)(1024 * V_LVD / VREF)) Flag |= 2;
	return (ad >= (uint16_t)(1024 * V_SDN / VREF)) ? 1 : 0;
}



/*---------------------------------------------*/
/* Read temperature from MAX31855              */
/*---------------------------------------------*/

int16_t get_temp (void)
{
	uint32_t d;
	uint16_t th;
	int8_t c;

	/* Read a 32-bit data from MAX31855 */
	PORTA &= ~_BV(3);	/* CS# = L */
	for (c = d = 0; c < 32; c++) {
		d <<= 1;
		if (PINA & _BV(2)) d |= 1;
		PORTA |= _BV(4);	/* CK = H */
		PORTA &= ~_BV(4);	/* CK = L */
	}
	PORTA |= _BV(3);	/* CS# = H */

	/* Fault? */
	if (d & 0x10000) return -1000;

	/* Get hot junction tempearture in unit of degree C */
	th = d >> 18;
	if (th & 0x2000) th |= 0xC000;
	return ((int16_t)th + 2) / 4;
}



/*---------------------------------------------*/
/* Display/Send a value                        */
/*---------------------------------------------*/

void disp_val (
	int val	/* Value (-999..1999) */
)
{
	static const uint8_t seg[] = {0x7E, 0x30, 0x6D, 0x79, 0x33, 0x5B, 0x5F, 0x70, 0x7F, 0x7B, 0x00, 0x01};
	uint8_t i, d, buf[8];


	if (val < -999 || val > 1999) {	/* Out of valid range? */
		xputs(PSTR("fault\n"));	/* Serial output */
		buf[0] = 0; buf[1] = 0x4F; buf[2] = buf[3] = 0x05;	/* "Err" */

	} else {	/* In the valid range */
		xprintf(PSTR("%d\n"), val);	/* Serial output */
		xsprintf((char*)buf, PSTR("%4d"), val);	/* Value --> String */
		for (i = 0; i < 4; i++) {		/* String --> Segment */
			d = buf[i];
			d = (d >= '0' && d <= '9') ? d - '0' : (d == '-') ? 11 : 10;
			buf[i] = seg[d];
		}
	}

	/* Move extra segments */
	if (buf[0] & 0x20)	buf[1] |= 0x80;	/* bc1000 */
	if (buf[0] & 0x01)	buf[3] |= 0x80;	/* g1000 */
	if (Flag & 2) 		buf[2] |= 0x80;	/* LOBAT */

	/* Update display buffer */
	for (i = 0; i < sizeof Lcd; i++)
		Lcd[i] = buf[1 + i];
}


/*---------------------------------------------*/
/* Program Entry Point                         */
/*---------------------------------------------*/

int main (void)
{
	int n;


	MCUSR = 0;
	ACSR = _BV(ACD);		/* Disable AC */
	PRR = 0b11011110;		/* Disable TWI, TIM2, USART1, TIM1, SPI and USART0 */

	PORTA = 0b01101100;		/* zuHLHuLL */
	DDRA =  0b00111011;
	DDRB =  0b11111111;		/* LLLLLLLL */
	DDRC =  0b11111111;		/* LLLLLLLL */
	DDRD =  0b11111111;		/* LLLLLLLL */

	/* Start Timer0 as 300Hz IVT (COMPA) */
	OCR0A =  F_CPU / 256 / 300 - 1;
	TCCR0A = _BV(WGM01);
	TCCR0B = 0b100;
	TIMSK0 = _BV(OCIE0A);

	set_sleep_mode(SLEEP_MODE_IDLE);
	sei();

	xdev_out(suart_putc);	/* Join suart_putc() to xitoa module as stdout */

	/* Lamp test */
	Lcd[0] = 0xFF; Lcd[1] = 0xFF; Lcd[2] = 0xFF;	/* All segment on */
	for (n = 0; n < 250; n++) sleep_mode();			/* 0.8 sec */
	do sleep_mode(); while (BTN_PWR);				/* Wait for button released */
	Flag = 0;

	/* Measureing loop */
	do {
		sleep_mode();
		if (Flag & 1) {
			Flag &= ~1;
			if (!check_fuel()) break;
			n = get_temp();
			disp_val(n);
		}
	} while (!BTN_PWR);

	/* Shutdown */
	Lcd[0] = 0x1D; Lcd[1] = 0x47; Lcd[2] = 0x47;	/* "off" */
	for (n = 0; n < 300; n++) sleep_mode();			/* 1 sec */
	PORTA &= ~_BV(5);	/* Power off */
	for (;;) ;
}

